library gen_server;

import 'dart:convert';
import 'dart:async';

import 'package:cryptography/cryptography.dart';

import './data_utils.dart';

Future<SecretBox> encryptMessage(dynamic secretKey, dynamic message) async {
  final algorithm = AesCtr.with256bits(
    macAlgorithm: Hmac.sha256(),
  );
  // final nonce = algorithm.newNonce();
  final secretBox = await algorithm.encrypt(
    message,
    secretKey: secretKey,
  );
  return secretBox;
}

Future<List<int>> decryptMessage(dynamic secretKey, dynamic message) async {
  final algorithm = AesCtr.with256bits(
    macAlgorithm: Hmac.sha256(),
  );
  final clearText = await algorithm.decrypt(
    message,
    secretKey: secretKey,
  );
  return clearText;
}

class Role {

  late X25519 algorithm;
  late SimpleKeyPair keyPair;
  late SimplePublicKey pubKey;
  late SecretKey? secretKey = null;

  Role() {
    this.algorithm = X25519();
  }

  bool get ready {
    return secretKey != null;
  }

  Future<void> initialize() async {
    this.keyPair = await this.algorithm.newKeyPair();
    this.pubKey = await this.keyPair.extractPublicKey();
    this.secretKey = null;
  }

  Future<void> generateSharedSecret(SimplePublicKey pubKey) async {
    this.secretKey = await this.algorithm.sharedSecretKey(
      keyPair: this.keyPair,
      remotePublicKey: pubKey,
    );
  }

  Future<String> encrypt(String text) async {
    final message = str2Data(text);
    final secretBox = await encryptMessage(this.secretKey!, message);
    return jsonEncode({
      'c': secretBox.cipherText,
      'n': secretBox.nonce,
      'm': secretBox.mac.bytes,
    });
  }

  Future<String> decrypt(String text) async {
    final data = jsonDecode(text);
    final secretBox = SecretBox(toListInt(data['c']),
      nonce: toListInt(data['n']),
      mac: Mac(toListInt(data['m']))
    );
    final message = await decryptMessage(this.secretKey!, secretBox);
    return data2Str(message);
  }

  dynamic toJson() {
    return {
      'p': pubKey.bytes,
    };
  }

  dynamic toInfoJson() {
    return {
      'state': ready ? 'ready' : 'init',
    };
  }

}

SimplePublicKey pubKeyFromData(List<int> bytes) {
  return SimplePublicKey(bytes, type: KeyPairType.x25519);
}

Future<void> main() async {
  // final algorithm = X25519();

  // // Alice chooses her key pair
  // final aliceKeyPair = await algorithm.newKeyPair();

  // // Alice knows Bob's public key
  // final bobKeyPair = await algorithm.newKeyPair();
  // final bobPublicKey = await bobKeyPair.extractPublicKey();

  // // Alice calculates the shared secret.
  // final sharedSecret = await algorithm.sharedSecretKey(
  //   keyPair: aliceKeyPair,
  //   remotePublicKey: bobPublicKey,
  // );
  // // Encrypt/Decrypt
  // final data = 'asdfg';
  // final message = str2Data(data);
  // var enc_data = await encryptMessage(sharedSecret, message);
  // print(enc_data);
  // var dec_data = await decryptMessage(sharedSecret, enc_data);
  // print(data2Str(dec_data) == data2Str(message));

  Role r1 = Role();
  await r1.initialize();
  Role r2 = Role();
  await r2.initialize();
  await r1.generateSharedSecret(r2.pubKey);
  await r2.generateSharedSecret(r1.pubKey);
  final sharedSecret = r2.secretKey;

  print(r1.toJson());

  print(r1.pubKey);
  print(r1.pubKey.bytes);
  print(r2.pubKey.bytes);
  // print(pubKeyFromData(r1.pubKey.bytes));

  // Encrypt/Decrypt
  final data = 'asdfg';
  final enc_data = await r1.encrypt(data);
  var dec_data = await r1.decrypt(enc_data);
  print("Encrypted: $enc_data");
  print("Decrypted: $dec_data");
  print(data == dec_data);
  // print(jsonEncode({
  //   'c': enc_data.cipherText,
  //   'n': enc_data.nonce,
  //   'm': enc_data.mac.bytes,
  // }));
  // print(SecretBox(enc_data.cipherText,
  //   nonce: enc_data.nonce,
  //   mac: Mac(enc_data.mac.bytes)
  // ));
  // final message = str2Data(data);
  // var enc_data = await encryptMessage(sharedSecret, message);
  // print(enc_data);
  // var dec_data = await decryptMessage(sharedSecret, enc_data);
  // print(data2Str(dec_data) == data2Str(message));

}
