library gen_server;

import 'dart:io';
import 'dart:core';
import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import './crypter.dart';
import './data_utils.dart';
import './op_codes.dart' as OpCodes;

const int HandshakeStateInit = 0;
const int HandshakeStateReady = 1;
const int HandshakeStatePeerReady = 2;
const EndSymbol = '\t';

class GenSocket {

	late Socket? socket = null;
	late String? host;
	late int? port;
	late String? id = null;
	late Function(GenSocket gen_socket, String data) onMessageHandler = (GenSocket gen_socket, String data) {};
	late Function() onBeforeHandshakeHandler = () {};
	late Function() onAfterHandshakeHandler = () {};
	late Function() onHandshakeCompleteHandler = () {};
	late Function(dynamic error) onErrorHandler = (dynamic error) {};
	late Function() onCloseHandler = () {};
	late Role role;
	bool isServerSocket = false;
	int handshakeState = HandshakeStateInit;
	List messages = [];

	String get remoteAddress {
		return '${socket?.remoteAddress.address}:${socket?.remotePort}';
	}

	bool get isHandshakingFin {
		return handshakeState != HandshakeStateInit;
	}

	GenSocket({Socket? socket, String? host, int? port, String? id}) {
		if(socket == null) {
			this.host = host;
			this.port = port;
			isServerSocket = false;
		} else {
			this.socket = socket;
			isServerSocket = true;
		}
		if(id != null) {
			this.id = id;
		}
	}

	Future<void> initialize() async {
		if(socket == null) {
			Socket socket = await Socket.connect(host!, port!);
			this.socket = socket;
		}
		handshakeState = HandshakeStateInit;
		role = Role();
		await role.initialize();
	}

	void onMessage(Function(GenSocket gen_socket, String data) onMessageHandler) {
		this.onMessageHandler = onMessageHandler;
	}

	void onBeforeHandshake(Function() onBeforeHandshakeHandler) {
		this.onBeforeHandshakeHandler = onBeforeHandshakeHandler;
	}

	void onAfterHandshake(Function() onAfterHandshakeHandler) {
		this.onAfterHandshakeHandler = onAfterHandshakeHandler;
	}

	void onConnect(Function() onHandshakeCompleteHandler) {
		this.onHandshakeCompleteHandler = onHandshakeCompleteHandler;
	}

	void onError(Function(dynamic error) onErrorHandler) {
		this.onErrorHandler = onErrorHandler;
	}

	void onClose(Function() onCloseHandler) {
		this.onCloseHandler = onCloseHandler;
	}

	void start() {
		socket?.listen(
			// message
			(Uint8List data) {
			  handle(data);
			},
			// errors
			onError: (error) {
				destroy();
				handleError(error);
			},
			// close
			onDone: () {
				destroy();
				handleClose();
			},
		);
	}

	Future<void> processMessage(String message, Function(String payload) callback) async {
		List<String> messageList = message.split(EndSymbol);
		for(int index = 0; index < messageList.length; index++) {
			String payload = messageList[index];
			if(payload.length > 0) {
				await callback(payload);
			}
		}
	}

	Future<void> _handleMessage(String message) async {
		if(isHandshakingFin) {
			var originMessage = await role.decrypt(message);
			if(isServerSocket && handshakeState == HandshakeStateReady) {
				var jsonData = jsonDecode(originMessage);
				switch(jsonData['a']) {
					case OpCodes.HandshakeComplete:
					handshakeState = HandshakeStatePeerReady;
					await onHandshakeComplete();
					break;
					default:
					// do nothing
					break;
				}
			} else {
				onMessageHandler(this, originMessage);
			}
		} else {
			var jsonData = jsonDecode(message);
			switch(jsonData['a']) {
				case OpCodes.Handshake:
				var pubKeyData = jsonData['p']['r']['p'];
				await onBeforeHandshakeHandler();
				await handshake(toListInt(pubKeyData));
				if(jsonData['p']['id'] != null) {
					final id = jsonData['p']['id'];
					this.id = id;
				}
				await onAfterHandshakeHandler();
				if(isServerSocket) {
					// do nothing
				} else {
					await onHandshakeComplete();
				}
				break;
				default:
				// do nothing
				break;
			}
		}
	}

	Future<void> onHandshakeComplete() async {
		await resendQueuedMessages();
		await onHandshakeCompleteHandler();
	}

	void handle(Uint8List data) async {
		var message = data2Str(data);
		await processMessage(message, (String payload) async {
			await _handleMessage(payload);
		});
	}

	void handleError(dynamic error) {
		this.onErrorHandler(error);
	}

	void handleClose() {
		this.onCloseHandler();
	}

	Future<void> _write(String message) async {
		try {
			if(isHandshakingFin) {
				var data = await role.encrypt(message);
				socket?.write(data + EndSymbol);
			} else {
				socket?.write(message + EndSymbol);
			}
		} catch(err) {
			print("Error::");
			print(err);
			_prependMessage(message);
		}
	}

	void _prependMessage(String message) {
		messages.insert(0, message);
	}

	void _appendMessage(String message) {
		messages.add(message);
	}

	Future<void> sendMessage(String message) async {
		if(isServerSocket) {
			if(handshakeState == HandshakeStatePeerReady) {
				await _write(message);
			} else {
				_appendMessage(message);
			}
		} else {
			if(handshakeState == HandshakeStateReady) {
				await _write(message);
			} else {
				_appendMessage(message);
			}
		}
	}

	Future<void> resendQueuedMessages() async {
		for(int index = 0; index < messages.length; index++) {
			String message = messages[index];
			await sendMessage(message);
		}
	}

	Future<void> flush() async {
		return await socket?.flush();
	}

	void destroy() {
		socket?.destroy();
	}

	Future<void> close() async {
		return await socket?.close();
	}

	Future<void> disconnect() async {
		await flush();
		await close();
		destroy();
	}

	Future<void> reconnect() async {
		socket = null;
		await initialize();
		start();
		await sendHandshakeInfo();
	}

	Future<void> handshake(List<int> pubKeyBytes) async {
		var pubKey = pubKeyFromData(pubKeyBytes);
		await role.generateSharedSecret(pubKey);
		handshakeState = HandshakeStateReady;
	}

	Future<void> sendHandshakeInfo() async {
		await _write(jsonEncode({
	      'a': OpCodes.Handshake,
	      'p': toJson(),
	    }));
	}

	Future<void> sendHandshakeComplete() async {
		await _write(jsonEncode({
	      'a': OpCodes.HandshakeComplete,
	    }));
	}

	dynamic toJson() {
		return {
			'r': role.toJson(),
			'id': id,
		};
	}

	dynamic toInfoJson() {
		return {
			'r': role.toInfoJson(),
		};
	}

}
